'''
Tests run independently on each other, but not in a single python interpreter.
After each test a fresh `python` is required (it won't run as casual suite),
so either run them like this one by one::

    python test_<something>.py

or use `KivyUnitTest <https://github.com/KeyWeeUsr/KivyUnitTest>`_ to do it
instead of you and better.
'''
